import numpy
SCENARIO_VERSION = '2020a'  # default scenario version for writing scenario files
SUPPORTED_COMMONROAD_VERSIONS = {'2018b', '2020a'}  # supported version for reading scenario files
TWO_PI = 2.0 * numpy.pi
